# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------
import threading
import time

from azdata.cli.core.clients.cli_client import CliClient
from azdata.cli.core.deploy import display
from azdata.cli.core.logging import get_logger
import azdata.cli.core.util as core_util

__all__ = ['beget']

CONNECTION_RETRY_ATTEMPTS = 12
RETRY_INTERVAL = 5
UPDATE_INTERVAL = (15 * 60) / RETRY_INTERVAL
logger = get_logger(__name__)

def beget(_):
    """Client factory"""
    return ConfigClientMixin(check_auth=True, check_eula=False)



class ConfigClientMixin(CliClient):
    def __init__(self, check_auth=False, check_eula=False):
        """
        :param check_auth: Set to `True` to have this command enforce authentication
                           before running.
        :param check_auth: Set to `True` to have this command enforce eula
                           acceptance before running.
        """
        super(ConfigClientMixin, self).__init__(check_auth=check_auth, check_eula=check_eula)

    def upgrade_config(self, description, abort=False):
        """
        Upgrade configuration settings.
        :param description
        :param abort
        :return:
        """

        display("Applying Big Data Cluster settings.")
        self.apis.controller.upgrade_config(description, abort)

        # Keep track of status thread to join in case of deployment error
        #
        stop = threading.Event()
        status_thread = threading.Thread(
            target=self.apis.control_deploy.get_controller_logs, args=[stop, False])
        status_thread.setDaemon(True)
        status_thread.start()

        cluster_state = self.apis.controller.get_cluster_state()

        # It can take a second or two for the state machine to move back into the upgrading state.
        # If the upgrade is still old, retry.
        #
        while cluster_state == "upgradepaused":
            time.sleep(RETRY_INTERVAL)

            cluster_state = self.apis.controller.get_cluster_state()

        i = 0
        while cluster_state != "ready":
            if cluster_state == "error":
                stop.set()
                status_thread.join()

                raise Exception("Cluster settings apply failed. Cluster State: {0}"
                                .format(cluster_state))
            elif cluster_state == "upgradepaused":
                display(
                    "The cluster settings apply has been paused. Please examine the Big Data Cluster "
                    "and either retry the settings apply with 'azdata bdc settings apply' "
                    "or cancel the settings apply by running 'azdata bdc settings cancel-apply'.")
                return

            if i !=0 and i % UPDATE_INTERVAL == 0:
                logger.warning("Cluster is not ready after {0} minutes."
                               .format(core_util.minutes_elapsed(i, RETRY_INTERVAL)))

            if cluster_state is not None:
                logger.debug("Waiting for cluster to be in Ready state. Current state : {0}"
                             .format(cluster_state))

            time.sleep(RETRY_INTERVAL)
            cluster_state = self.apis.controller.get_cluster_state()
            i = i + 1

        display("The cluster settings have been applied successfully. To view the status, please run 'azdata bdc status show'.")

    def set_config(self, service, resource, settings):
        """
        Sets the configuration settings for given service and resource
        :param service: The name of the service for the settings
        :param resource: The name of the resource for the settings
        :param settings: A dictionary of setting key/value.
        :return:
        """
        self.apis.controller.set_config(service, resource, settings)
        display("Settings for service '%s' and resource '%s' are set." % (service, resource))

    def show_config(self, service, resource, settings, filter_option, recursive, include_details):
        """
        Gets the configuration settings for given service and resource
        :param service: The name of the service for the settings
        :param resource: The name of the resource for the settings
        :param filter_option: The option to filter the settings.
        :param recursive: The flag to denote if settings have to be shown recursive.
        :param include_details : The flag to denote if status has to be included in response.
        :return:
        """
        return self.apis.controller.show_config(service, resource, settings, filter_option, recursive, include_details)

    def revert_pending_settings(self):
        """
        Revert pending configuration settings.
        :return:
        """
        display("Reverting pending settings...")
        self.apis.controller.revert_config_settings()
