# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------
import json
import re

from azdata.cli.core.exceptions import CliError, ApiError, ControllerError
from azdata.cli.commands.config.util import hierarchical_output
from azdata.cli.core.prompt import prompt_y_n

## TODO: Service names are also defined in azdata.cli.commands.constants.
## either move all of them to core or comebine those definition.
##
from azdata.cli.core.util import raise_command_error
from knack.util import CommandResultItem

CONTROL_SERVICE = 'control'
SQL_SERVER_SERVICE = 'sql'
HDFS_SERVICE = 'hdfs'
SPARK_SERVICE = 'spark'
GATEWAY_SERVICE = 'gateway'
APP_SERVICE = 'app'
"""
Names of services
"""


def config_upgrade(client, force=False):
    """
    Upgrades configuration in the cluster.
    :param client: Minxin client
    :param description: The optional description for this configuration upgrade
    :param force: boolean to indicate whether force upgrade without confirmation
    :return:
    """
    # TODO: add description in command options when we are ready to expose
    # description to customer
    description = ""

    if not force:
        get_pending_settings(client)
        yes = prompt_y_n('Do you want to continue applying the pending settings?')
    else:
        yes = force  # yes|no for no-tty

    if yes is not True:
        msg = 'Exiting settings apply due to user input. ' + \
              'Settings will not be applied at this time.'
        return msg
    try:
        client.stdout('NOTE: Applying pending settings will cause a restart of services.')
        client.stdout('')
        client.upgrade_config(description)
    except ControllerError as e:
        raise CliError(e)
    except Exception as e:
        raise CliError(e)


def config_abort_upgrade(client, force=False):
    """
    Aborts the configuration upgrade in the cluster.
    :param client: Minxin client
    :param force: boolean to indicate whether force upgrade without confirmation
    :return:
    """
    # TODO: add description in command options when we are ready to expose
    # description to customer
    description = ""

    if not force:
        yes = prompt_y_n('Do you want to continue cancelling the pending settings application?')
    else:
        yes = force  # yes|no for no-tty

    if yes is not True:
        msg = 'Exiting cancelling of settings apply due to user input.' + \
              'Settings apply will not be cancelled at this time.'
        return msg
    try:
        client.stdout('NOTE: Settings cancel-apply will cause restart of services.')
        client.stdout('')
        client.upgrade_config(description, True)
    except ControllerError as e:
        raise CliError(e)
    except Exception as e:
        raise CliError(e)


def bdc_sql_config_set(client, resources=None, settings=None):
    """
    Sets the configuration settings for sql service
    """
    return bdc_config_set_helper(client, SQL_SERVER_SERVICE, resources, settings)


def bdc_hdfs_config_set(client, resources=None, settings=None):
    """
    Sets the configuration settings for HDFS service
    """
    return bdc_config_set_helper(client, HDFS_SERVICE, resources, settings)


def bdc_spark_config_set(client, resources=None, settings=None):
    """
    Sets the configuration settings for spark service
    """
    return bdc_config_set_helper(client, SPARK_SERVICE, resources, settings)


def bdc_gateway_config_set(client, resources=None, settings=None):
    """
    Sets the configuration settings for gateway service
    """
    return bdc_config_set_helper(client, GATEWAY_SERVICE, resources, settings)


def bdc_app_config_set(client, resources=None, settings=None):
    """
    Sets the configuration settings for app service
    """
    return bdc_config_set_helper(client, APP_SERVICE, resources, settings)


def bdc_config_set(client, settings):
    """
    Sets the configuration settings for cluster
    """
    return bdc_config_set_helper(client, None, None, settings)


def bdc_config_set_helper(client, service, resources, settings):
    """
    Parse settings from list into dictionary, and call client to set_config
    """
    try:
        setting_dict = {}
        if not settings:
            raise CliError("No settings are specified.")

        # Split the settings string by comma which is followed by a key [\w.-]+ and equal sign.
        # [\w.-]+ includes characters from a-z, A-Z, 0-9, underscore, dot and minus character.
        # Assuming the setting string is "key1=value1,key2=value2...", we will get ["key1=value1", "key2=value2"...]
        #
        setting_list = re.split(',(?=[\w.-]+=)', settings.strip())

        for setting in setting_list:
            # The setting key won't contain equal sign. So the first equal sign split key and value.
            #
            setting_key_value_pair = setting.strip().split("=", 1)
            if len(setting_key_value_pair) != 2:
                raise CliError("Please specify the key value pair separated by equal sign like '<key1>=<value1>,<key2>=<value2>...'")
            setting_dict[setting_key_value_pair[0].strip()] = setting_key_value_pair[1].strip()

        if len(setting_dict) == 0:
            raise CliError("No settings are specified.")

        if resources != None:
            resource_list = resources.strip().split(",")
            for resource in resource_list:
                client.set_config(service, resource.strip(), setting_dict)
        else:
            client.set_config(service, None, setting_dict)
        client.stdout("Successfully set settings.")
        return
    except ControllerError as e:
        raise CliError(e)
    except Exception as e:
        raise CliError(e)


def bdc_sql_settings_show(
        client,
        resources=None,
        settings=None,
        filter_option="userConfigured",
        recursive=False,
        include_details=False,
        description=False):
    """
    Gets the configuration settings for sql service
    """
    return bdc_config_show_helper(
        client,
        SQL_SERVER_SERVICE,
        resources,
        settings,
        filter_option,
        recursive,
        include_details,
        description)


def bdc_hdfs_settings_show(
        client,
        resources=None,
        settings=None,
        filter_option="userConfigured",
        recursive=False,
        include_details=False,
        description=False):
    """
    Gets the configuration settings for hdfs service
    """
    return bdc_config_show_helper(
        client,
        HDFS_SERVICE,
        resources,
        settings,
        filter_option,
        recursive,
        include_details,
        description)


def bdc_spark_settings_show(
        client,
        resources=None,
        settings=None,
        filter_option="userConfigured",
        recursive=False,
        include_details=False,
        description=False):
    """
    Gets the configuration settings for spark service
    """
    return bdc_config_show_helper(
        client,
        SPARK_SERVICE,
        resources,
        settings,
        filter_option,
        recursive,
        include_details,
        description)


def bdc_gateway_settings_show(
        client,
        resources=None,
        settings=None,
        filter_option="userConfigured",
        recursive=False,
        include_details=False,
        description=False):
    """
    Gets the configuration settings for gateway service
    """
    return bdc_config_show_helper(
        client,
        GATEWAY_SERVICE,
        resources,
        settings,
        filter_option,
        recursive,
        include_details,
        description)


def bdc_app_settings_show(
        client,
        resources=None,
        settings=None,
        filter_option="userConfigured",
        recursive=False,
        include_details=False,
        description=False):
    """
    Gets the configuration settings for app service
    """
    return bdc_config_show_helper(
        client,
        APP_SERVICE,
        resources,
        settings,
        filter_option,
        recursive,
        include_details,
        description)


def bdc_settings_show(
        client,
        settings=None,
        filter_option="userConfigured",
        recursive=False,
        include_details=False,
        description=False):
    """
    Gets the configuration settings for cluster
    """
    return bdc_config_show_helper(
        client,
        None,
        None,
        settings,
        filter_option,
        recursive,
        include_details,
        description)


def bdc_config_show_helper(
        client,
        service,
        resources,
        settings,
        filter_option,
        recursive,
        include_details,
        description=False):
    """
    Parse comma separated list of settings into a list, and call client to show_config
    """
    try:
        # --description argument can be used only with --include_details.
        #
        if description:
            if not include_details:
                raise ValueError('[--description] can be used only with [--include-details].')
        settings_list = []
        if settings:
            settings_list = settings.strip().split(",")

        if resources:
            resource_list = resources.strip().split(",")
            resources = list()
            has_settings = False
            for resource in resource_list:
                response = client.show_config(
                    service, resource, settings_list, filter_option, recursive, include_details)
                resources.append(response)

                if bool(response["settings"]):
                    has_settings = True

            # Fix for API discrepancy
            if has_settings:
                result = dict()
                result["serviceName"] = service
                result["settings"] = {}
                result["resources"] = resources
            else:
                result = resources

        else:
            response = client.show_config(service, None, settings_list, filter_option, recursive, include_details)
            result = response

        if description:
            result["description"] = description

        return result

    except ControllerError as e:
        raise CliError(e)
    except Exception as e:
        raise CliError(e)


def config_revert(client, force=False):
    """
    Revert pending configuration settings in the cluster.
    :param client: Mixin client
    :param force: boolean to indicate whether force revert without confirmation
    :return:
    """
    if not force:
        get_pending_settings(client)
        yes = prompt_y_n('Do you want to continue with reverting pending settings?')
    else:
        yes = force  # yes|no for no-tty

    if not yes:
        msg = 'Pending settings were not reverted.'
        return msg
    try:
        client.revert_pending_settings()
        client.stdout("Successfully reverted pending settings.")
    except ControllerError as e:
        raise CliError(e)
    except Exception as e:
        raise CliError(e)


def get_pending_settings(client):
    """
    Get pending settings in all scopes of the cluster.
    """
    client.stdout("Getting pending settings...")
    response = bdc_config_show_helper(client, None, None,
                                      settings=None,
                                      filter_option="pending",
                                      recursive=True,
                                      include_details=True,
                                      description=False)
    commandResult = CommandResultItem(response)
    client.stdout(hierarchical_output(commandResult))
    return response
