# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

from collections import OrderedDict
import textwrap


CHAR_LIMIT = 20
"""
Character limit for text wrapping on the hierarchical output.
"""


def order_config(section, description=False):
    """
    Order config `dict` sections to the same order the server API handed us.
    NOTE: This is redundant in Python 3.7 however needed for earlier versions.

    :param section: The config `dict` section to order.
    :param description: Whether the description property should be included.
    :return: A well defined `OrderedDict` of the given config section.
    """

    def get_settings(settings, status=None):
        """
        Creates ordered dictionaries for the given settings to be used in the BoxLayout.
        :param settings:
        :param status:
        :return:
        """
        def new_setting(name, value, stat):
            if stat is None:
                return OrderedDict([
                    ("name", textwrap.fill(name, CHAR_LIMIT)),
                    ("running", textwrap.fill(value, CHAR_LIMIT)),
                    ("options", [])
                ])
            elif description:
                return OrderedDict([
                    ("name", textwrap.fill(name, CHAR_LIMIT)),
                    ("description", textwrap.fill(stat["description"], CHAR_LIMIT)),
                    ("default", textwrap.fill(stat["defaultValue"], CHAR_LIMIT)),
                    ("running", textwrap.fill(value, CHAR_LIMIT)),
                    ("configured", textwrap.fill(stat["configuredValue"], CHAR_LIMIT)),
                    ("last updated", stat["lastUpdatedTime"]),
                    ("configurable", stat["configurable"]),
                    ("immutable", stat["immutable"]),
                    ("options", [])
                ])
            else:
                return OrderedDict([
                    ("name", textwrap.fill(name, CHAR_LIMIT)),
                    ("default", textwrap.fill(stat["defaultValue"], CHAR_LIMIT)),
                    ("running", textwrap.fill(value, CHAR_LIMIT)),
                    ("configured", textwrap.fill(stat["configuredValue"], CHAR_LIMIT)),
                    ("last updated", stat["lastUpdatedTime"]),
                    ("configurable", stat["configurable"]),
                    ("immutable", stat["immutable"]),
                    ("options", [])
                ])

        if status is None:
            return [new_setting(name, value, None) for name, value in settings.items()]
        else:
            return [new_setting(stat["settingName"], settings[stat["settingName"]], stat) for stat in status]

    def get_resources(obj):
        """
        Returns the configuration settings of all resources.
        :param obj:
        :return:
        """
        def new_resource(o):
            """
            Creates a new ordered dictionary resource object to be used int he BoxLayout.
            :param o:
            :return:
            """
            return None if "resourceName" not in o else OrderedDict([
                ("resourceName", o["resourceName"]),
                ("settings", get_settings(o["settings"], o["status"]) if o.get("status")
                    else get_settings(o["settings"])),
            ])

        obj = obj if obj else []
        if isinstance(obj, dict):
            return new_resource(obj)

        return [new_resource(resource) for resource in obj]

    def get_services(obj):
        """
        Returns the configuration settings of all services.
        :param obj:
        :return:
        """
        def new_service(o):
            """
            Creates a new ordered dictionary service object to be used in the BoxLayout.
            Only adds resources if those resources have listed settings.
            :param o:
            :return:
            """
            if "serviceName" not in o:
                return None
            else:
                # Only list the resources that have settings in the tabular format
                result = get_resources(o.get("resources"))
                resources = list()
                for r in result:
                    if r["settings"]:
                        resources.append(r)

                return OrderedDict([
                    ("serviceName", o["serviceName"]),
                    ("settings", get_settings(o["settings"], o["status"]) if o.get("status")
                        else get_settings(o["settings"])),
                    ("resource", resources)
                ])

        obj = obj if obj else []
        if isinstance(obj, dict):
            return new_service(obj)

        return [new_service(service) for service in obj]

    def get_bdc(obj):
        """
        Retrieves all configuration settings in an ordered dictionary for the given BDC to be used in the BoxLayout.
        :param obj:
        :return:
        """
        return None if "bdcName" not in obj else OrderedDict([
            ("clusterName", obj["bdcName"]),
            ("settings", get_settings(obj["settings"], obj["status"]) if obj.get("status")
                else get_settings(obj["settings"])),
            ("service", get_services(obj.get("services")))
        ])

    return {
        "bdc": get_bdc,
        "service": get_services,
        "resource": get_resources
    }.get(section)


def hierarchical_output(command_result):
    """
    Callback for formatting complex custom-output.
    :parm_am command_result: The command's high-level result object.
    :return: Complex BoxLayout otherwise flat json.
    """
    from azdata.cli.core.layout import BoxLayout

    try:
        raw_result = command_result.result
        result = order_config('bdc', raw_result.get('description'))(raw_result)
        result = result if result else order_config('service', raw_result.get('description'))(raw_result)
        result = result if result else order_config('resource', raw_result.get('description'))(raw_result)

        return BoxLayout(result, config={
            'headers': {
                'left': {
                    'label': '',
                    'id': None
                },
                'right': {
                    'label': '',
                    'id': None
                }
            },
            'identifiers': []
        }, bdc_config=True)
    except Exception as e:  # -- fallback --
        from knack.output import format_json
    return format_json(command_result)
