# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------
import json
import time
import os
from azdata.cli.core.logging import get_logger
from azdata.cli.commands.spark.client import SparkClientMixin
from azdata.cli.commands.spark.custom import (
    spark_batch_create,
    spark_batch_info,
    spark_batch_list,
    spark_batch_log,
    spark_batch_state,
)

__all__ = ["beget"]

def _get_knox_username():
    # default to root if SESSION_AZDATA_USERNAME is not set.
    user = os.getenv('DISTCP_AZDATA_USERNAME',os.getenv('AZDATA_USERNAME','root'))

    return user

def _get_knox_password():
    # check for SESSION_AZDATA_PASSWORD else default to AZDATA_PASSWORD
    password = os.getenv('DISTCP_AZDATA_PASSWORD',os.getenv('AZDATA_PASSWORD',None))

    return password
def beget(_):
    """Client factory"""
    return DistcpClientMixin(check_auth=False, check_eula=False)


logger = get_logger(__name__)


class DistcpClientMixin(SparkClientMixin):
    def __init__(self, check_auth=True, check_eula=True):
        self.file_to_execute = self.get_distcp_file_path()
        self.class_name = "com.microsoft.sqlserver.SparkDistCp"
        super(DistcpClientMixin, self).__init__(
            check_auth=check_auth, check_eula=check_eula
        )

    def get_distcp_file_path(self):
        """
        Returns the distcp symbolic link jar complete path residing on the nodemanager
        """
        path = "local:/opt/hadoop/share/hadoop/tools/lib/hadoop-distcp.jar"
        return path

    def prepare_arguments(
        self,
        from_path,
        to_path,
        preserve=None,
        ignoreFailure=None,
        log=None,
        maps=None,
        overwrite=None,
        update=None,
        append=None,
        filters=None,
        delete=None,
        strategy=None,
        bandwidth=None,
        atomic=None,
        diff=None,
        rdiff=None,
        numListstatusThreads=None,
        skipcrccheck=None,
        blocksperchunk=None,
        copybuffersize=None,
        xtrack=None,
        direct=None,
        configuration=None,
    ):
        """
        Prepare the arguments to for the spark-sumbit job.
        :param from_path
        :param to_path
        :param preserve
        :param ignoreFailure
        :param log
        :param maps
        :param overwrite
        :param update
        :param append
        :param filters
        :param delete
        :param strategy
        :param bandwidth
        :param atomic
        :param diff
        :param rdiff
        :param numListstatusThreads
        :param skipcrccheck
        :param blocksperchunk
        :param copybuffersize
        :param xtrack
        :param direct
        :param configuration
        """
        argumentsArray = []
        argumentsArray.append("-Dipc.client.fallback-to-simple-auth-allowed=true")
        argumentsArray.append("-Ddfs.webhdfs.api.use.delegationtoken=false")
        argumentsArray.append("-Ddfs.webhdfs.encryption.zone.enabled=false")

        if configuration is not None:
            config_params = self._decode_dictionary(configuration)
            for attribute, value in config_params.items():
                argumentsArray.append("-D" + attribute.strip() + "=" + value.strip())
        if preserve is not None:
            argumentsArray.append("-p " + preserve)
        if ignoreFailure is not None:
            argumentsArray.append("-i")
        if maps is not None:
            argumentsArray.append("-m " + maps)
        if overwrite is not None:
            argumentsArray.append("-overwrite")
        if update is not None:
            argumentsArray.append("-update")
        if append is not None:
            argumentsArray.append("-append")
        if delete is not None:
            argumentsArray.append("-delete")
        if strategy is not None:
            argumentsArray.append("-strategy " + str(strategy))
        if bandwidth is not None:
            argumentsArray.append("-bandwidth " + bandwidth)
        if atomic is not None:
            argumentsArray.append("-atomic")
        if diff is not None:
            argumentsArray.append("-diff")
        if numListstatusThreads is not None:
            argumentsArray.append("-numListstatusThreads " + numListstatusThreads)
        if skipcrccheck is not None:
            argumentsArray.append("-skipcrccheck")
        if blocksperchunk is not None:
            argumentsArray.append("-blocksperchunk " + blocksperchunk)
        if copybuffersize is not None:
            argumentsArray.append("-copybuffersize " + copybuffersize)
        if xtrack is not None:
            argumentsArray.append("-xtrack " + str(xtrack))
        if direct is not None:
            argumentsArray.append("-direct")

        argumentsArray.append(from_path)
        argumentsArray.append(to_path)
        if filters is not None:
            argumentsArray.append("-filters " + str(filters))
        if log is not None:
            argumentsArray.append("-log " + log)
        return json.dumps(argumentsArray)

    def job_copy(
        self,
        from_path,
        to_path,
        preserve=None,
        ignoreFailure=None,
        log=None,
        maps=None,
        overwrite=None,
        update=None,
        append=None,
        filters=None,
        delete=None,
        strategy=None,
        bandwidth=None,
        atomic=None,
        diff=None,
        rdiff=None,
        numListstatusThreads=None,
        skipcrccheck=None,
        blocksperchunk=None,
        copybuffersize=None,
        xtrack=None,
        direct=None,
        configuration=None,
    ):
        """
        Prepare the argument and submit the spark job to perform data copy job.
        :param from_path
        :param to_path
        :param preserve
        :param ignoreFailure
        :param log
        :param maps
        :param overwrite
        :param update
        :param append
        :param filters
        :param delete
        :param strategy
        :param bandwidth
        :param atomic
        :param diff
        :param rdiff
        :param numListstatusThreads
        :param skipcrccheck
        :param blocksperchunk
        :param copybuffersize
        :param xtrack
        :param direct
        :param configuration
        """
        config = {}
        config["spark.driver.extraClassPath"] = "/opt/hadoop/share/hadoop/tools/lib/hadoop-distcp.jar"
        config["spark.executor.extraClassPath"] = "/opt/hadoop/share/hadoop/tools/lib/hadoop-distcp.jar"
        config["spark.yarn.maxAppAttempts"] = "1"
        config["spark.executor.instances"] = "1"

        # get the knox_username and knowx_password for non secure cluster
        knox_user = _get_knox_username()
        knox_password = _get_knox_password()

        #if both are not None, set the environment variable.
        if knox_user is not None and knox_password is not None:
          config["spark.yarn.appMasterEnv.HADOOP_USER_NAME"] = knox_user
          config["spark.yarn.appMasterEnv.HADOOP_USER_PASSWORD"] = knox_password

        # check if ad user and keytab are set
        ad_username, keytab_hdfs_path = self.get_ad_username_keytab()

        if ad_username is not None and keytab_hdfs_path is not None:
            config["spark.yarn.appMasterEnv.KRB5PRINCIPAL"] = ad_username
            config["spark.yarn.appMasterEnv.KRB5KEYTAB"] = "./{}".format(os.path.basename(keytab_hdfs_path))

        argumentsJson = self.prepare_arguments(
            from_path,
            to_path,
            preserve,
            ignoreFailure,
            log,
            maps,
            overwrite,
            update,
            append,
            filters,
            delete,
            strategy,
            bandwidth,
            atomic,
            diff,
            rdiff,
            numListstatusThreads,
            skipcrccheck,
            blocksperchunk,
            copybuffersize,
            xtrack,
            direct,
            configuration,
        )
        name = "distcp_" + str(int(time.time()))
        logger.debug(
            "Running distcp: arguments {0}, name '{1}'".format(
                str(argumentsJson), name
            )
        )
        result = spark_batch_create(
            client=self,
            file_to_execute=self.file_to_execute,
            arguments=argumentsJson,
            class_name=self.class_name,
            name=name,
            files=keytab_hdfs_path,
            configuration=json.dumps(config)
        )
        return result

    def get_ad_username_keytab(self):
        username = self.profile.active_context.username
        keytabPath = os.getenv('DISTCP_KRB5KEYTAB',None)
        return username, keytabPath

    def job_info(self, job_id):
        """
        Get the status information for the provided job-id
        """
        return spark_batch_info(client=self, batch_id=job_id)

    def job_list(self):
        """
        List all running distcp job name starts with 'distcp_'
        """
        response = spark_batch_list(client=self)
        sessions = response.get('sessions', None)
        if sessions != None and len(sessions) > 0:
            newlist = []
            total = 0
            for obj in sessions:
                if str(obj['name']).startswith('distcp_'):
                    newlist.append(obj)
                    total+=1
            response['sessions'] = newlist
            response['total'] = total
        return response

    def job_log(self, job_id):
        """
        Get detailed log of running distcp job
        """
        return spark_batch_log(client=self, batch_id=job_id)

    def job_state(self, job_id):
        """
        Get the state of distcp job
        """
        return spark_batch_state(client=self, batch_id=job_id)

    def _decode_dictionary(self, value):
        """
        Some MapReduce parameters need to be dictionary of name value pairs.  For example configuration dictionary.
        This function checks to see if the content is a json dictionary and if so decodes it.

        Example:
            { "name": "value", "another_name": "another_value" }
        """
        if value.strip().startswith('{'):
            result = json.loads(value)
        else:
            raise ValueError("Value must be JSON encoded dictionary [{0}].".format(value))

        return result