# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

import os
import urllib3
import posixpath
from azdata.cli.core.enums import SecurityStrategy
from azdata.cli.core.clients.cli_client import CliClient
from azdata.cli.core.exceptions import CliError
from azdata.cli.core.enums import ContextType
from azdata.cli.core.extwebhdfs import ExtendedPyWebHdfsClient
from azdata.cli.commands.hdfsencryptionzone.exceptions import *

__all__ = ["beget"]

# By default urllib3 prints a warning message to the console if using SSL and server certificate
# validation is not enabled.  This disables the warning.
urllib3.disable_warnings()

def beget(_):
    """Client factory"""
    return HdfsencryptionzoneClientMixin(check_auth=True, check_eula=True)

class HdfsencryptionzoneClientMixin(CliClient):
    def __init__(self, check_auth=True, check_eula=True):
        """
        :param check_auth: Set to `True` to have this command enforce
                           authentication before running.
        :param check_eula: Set to `True` to have this command enforce eula
                           acceptance before running.
        """
        super(HdfsencryptionzoneClientMixin, self).__init__(
            check_auth=check_auth, check_eula=check_eula
        )

    def get_knox_endpoint(self):
        """
        Call the controller to get the webhdfs endpoint URL
        """
        try:
            endpoint = self.apis.controller.get_endpoints('webhdfs')['endpoint']
            if len(endpoint) == 0:
                raise CliError("No webhdfs endpoint was found")

            # The endpoint must end with '/' in order for the APIs to succeed
            if not endpoint[len(endpoint) - 1] == '/':
                endpoint = endpoint + '/'
        except Exception as e:
            raise CliError(e)

        return endpoint

    def login(self, knox_url, username=None, password=None):
        """
        Login to the webhdfs endpoint
        """
        self.hdfs = ExtendedPyWebHdfsClient(
            base_uri_pattern=knox_url
        )

        if self.profile.active_context.security_strategy == SecurityStrategy.basic:
            self.hdfs.enable_basic_auth(username, password)
        elif self.profile.active_context.security_strategy == SecurityStrategy.ad:
            self.hdfs.enable_kerberos_auth()
        else:
            raise CliError("Invalid security strategy.  Use azdata login to choose supported strategy.")

        return

    def create_encryption_zone(self, folder_path, key_name):
        """
        Set the folder to an encryption zone using the key name.
        The folder should exist and the key should have been created.
        """
        return self.hdfs.create_encryption_zone(folder_path, key_name)

    def list_encryption_zones(self):
        """
        Set the folder to an encryption zone using the key name.
        The folder should exist and the key should have been created.
        """
        return self.hdfs.list_encryption_zones()

    def get_file_encryption_info(self, file_path):
        """
        Get the encryption info for the given file.
        """
        return self.hdfs.get_file_encryption_info(file_path)

    def list_reencryption_status(self):
        """
        List the reencryption status for the encryption zones
        """
        return self.hdfs.list_reencryption_status()

    def reencrypt_encryption_zone(self, folder_path, action):
        """
        Perform the requested action for encryption zone reencryption.
        """
        return self.hdfs.reencrypt_encryption_zone(folder_path, action)