# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

import os
from azdata.cli.commands.hdfsencryptionzone.exceptions import HdfsencryptionzoneError
from azdata.cli.core.exceptions import CliError
from azdata.cli.core.logging import get_logger
from azdata.cli.core.enums import SecurityStrategy
from azdata.cli.core.prompt import prompt_pass
from azdata.cli.core.util import raise_command_error
from knack.prompting import NoTTYException
from azdata.cli.commands.hdfsencryptionzone.util import FormatEncryptionZones

logger = get_logger(__name__)

def _get_knox_username():
    # Note for now the knox username is hard coded to 'root'.
    return "root"


def _get_knox_password(client):
    # Note for now if the environment variable isn't set prompt for the password..
    if not os.getenv('AZDATA_PASSWORD'):
        try:
            client.stdout(
                "The Knox password is needed.  It can be set for the session by setting the environment variable "
                "AZDATA_PASSWORD.")
            password = prompt_pass('Enter Knox Password: ')
        except NoTTYException:
            raise NoTTYException('Missing AZDATA_PASSWORD environment variable in non-interactive environment.')
    else:
        password = os.environ['AZDATA_PASSWORD']

    return password

def _login(client):
    endpoint = client.get_knox_endpoint()

    if client.profile.active_context.security_strategy == SecurityStrategy.basic:
        username = _get_knox_username()
        password = _get_knox_password(client)

        client.login(endpoint, username, password)
    elif client.profile.active_context.security_strategy == SecurityStrategy.ad:
        client.login(endpoint)
    else:
        raise CliError("Invalid security strategy.  Use azdata login to choose supported strategy.")

    return

def hdfs_ez_create(client, folder_path, key_name):
    try:
        _login(client)
        result = client.create_encryption_zone(folder_path, key_name)
        return result
    except Exception as ex:
        raise_command_error(ex)


def hdfs_ez_list(client):
    try:
        _login(client)
        result = client.list_encryption_zones()
        return FormatEncryptionZones.format_encryption_zones(result['EncryptionZones'])
    except Exception as ex:
        raise_command_error(ex)

def hdfs_ez_get_file_encryption_info(client, file_path):
    try:
        _login(client)
        return client.get_file_encryption_info(file_path)
    except Exception as ex:
        raise_command_error(ex)

def hdfs_ez_reencrypt(client, folder_path, action):
    try:
        action_list = ['start', 'cancel']
        if not action.lower() in action_list:
            raise ValueError('Argument [--action] have valid values: start or cancel.')
        _login(client)
        return client.reencrypt_encryption_zone(folder_path, action)
    except Exception as ex:
        raise_command_error(ex)

def hdfs_ez_listreencryptionstatus(client):
    try:
        _login(client)
        result = client.list_reencryption_status()
        return FormatEncryptionZones.format_zone_reencryption_status(result['ZoneReencryptionStatuses'])
    except Exception as ex:
        raise_command_error(ex)