# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

import os
import json
import requests
from requests.auth import HTTPBasicAuth
import urllib3
import urllib.parse
from requests_kerberos import HTTPKerberosAuth, REQUIRED, OPTIONAL
import posixpath
from azdata.cli.core.enums import SecurityStrategy
from azdata.cli.core.clients.cli_client import CliClient
from azdata.cli.core.exceptions import CliError
from azdata.cli.core.enums import ContextType
from six.moves import http_client
from datetime import datetime as dt

__all__ = ["beget"]


def beget(_):
    """Client factory"""
    return HdfskeyClientMixin(check_auth=True, check_eula=True)


class HdfskeyClientMixin(CliClient):
    def __init__(self, check_auth=False, check_eula=False):
        """
        :param check_auth: Set to `True` to have this command enforce
                           authentication before running.
        :param check_auth: Set to `True` to have this command enforce eula
                           acceptance before running.
        """
        super(HdfskeyClientMixin, self).__init__(
            check_auth=check_auth, check_eula=check_eula
        )

    def get_hadoop_kms_endpoint(self):
        """
        Call the controller to get the webhdfs endpoint URL
        """
        try:
            endpoint = self.apis.controller.get_endpoints("hadoopkms")[
                "endpoint"
            ]

            # The endpoint must end with '/' in order for the APIs to succeed
            if not endpoint[len(endpoint) - 1] == "/":
                endpoint = endpoint + "/"
        except Exception as e:
            raise CliError(e)

        return endpoint

    def login(self, hadoopkms_url, username=None, password=None):
        """
        Login to the hadoopkms endpoint
        """
        self.hadoopkms_url_prefix = hadoopkms_url

        if (
            self.profile.active_context.security_strategy
            == SecurityStrategy.basic
        ):
            raise CliError(
                "Hadoop encryption zone key management is not available without Active Directory authentication."
            )
        elif (
            self.profile.active_context.security_strategy == SecurityStrategy.ad
        ):
            self.hadoopkms_auth = HTTPKerberosAuth(
                mutual_authentication=OPTIONAL,
                sanitize_mutual_error_response=False,
                force_preemptive=True,
            )
        else:
            raise CliError(
                "Invalid security strategy. Use azdata login to choose supported strategy."
            )
        self.request_extra_opts = {"verify": False, "auth": self.hadoopkms_auth}
        return

    def create_hdfskey(self, name, size=256):
        """
        Create a hdfs key with given key name.
        """
        uri = urllib.parse.urljoin(self.hadoopkms_url_prefix, "keys")
        response = self._submit_request(
            uri=uri,
            request_function=requests.post,
            allow_redirect=False,
            extra_headers={"content-type": "application/json"},
            body_data={"name": "{0}".format(name), "length": size},
        )
        return response.json()

    def list_hdfskey(self):
        """
        Create a hdfs key with given key name.
        """
        uri = urllib.parse.urljoin(self.hadoopkms_url_prefix, "keys/names")
        response = self._submit_request(
            uri=uri, request_function=requests.get, allow_redirect=False
        )

        if not response.status_code == http_client.OK:
            raise CliError(
                "list_hdfskey failed with error code : "
                + str(response.status_code)
            )
        return response.json()

    def roll_hdfskey(self, name):
        """
        Roll over a hdfs key with given key name.
        """
        uri = urllib.parse.urljoin(
            self.hadoopkms_url_prefix, "key/{0}".format(name)
        )
        response = self._submit_request(
            uri=uri,
            request_function=requests.post,
            allow_redirect=False,
            extra_headers={"content-type": "application/json"},
            body_data={},
        )
        return response.json()

    def describe_key(self, name):
        """
        Describe an hdfs key with a given name
        """
        uri = urllib.parse.urljoin(
            self.hadoopkms_url_prefix, "key/{0}/_metadata".format(name)
        )
        response = self._submit_request(
            uri=uri,
            request_function=requests.get,
            allow_redirect=False,
            extra_headers={"content-type": "application/json"},
            body_data={},
        )
        return response.status_code, response.json()

    def _submit_request(
        self,
        uri,
        request_function,
        allow_redirect,
        body_data=None,
        extra_headers={},
    ):
        extra_opts = self.request_extra_opts
        if not body_data == None:
            extra_opts["data"] = json.dumps(body_data)
        response = request_function(
            uri,
            allow_redirects=allow_redirect,
            headers=extra_headers,
            **extra_opts,
        )
        return response

