# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

import os
from azdata.cli.commands.hdfskey.exceptions import HdfskeyError
from azdata.cli.core.enums import SecurityStrategy
from azdata.cli.core.prompt import prompt_pass
from azdata.cli.core.exceptions import CliError
from azdata.cli.core.logging import get_logger
from azdata.cli.core.util import raise_command_error
from knack.prompting import NoTTYException

logger = get_logger(__name__)


def _login(client):
    endpoint = client.get_hadoop_kms_endpoint()

    if (
        client.profile.active_context.security_strategy
        == SecurityStrategy.basic
    ):
        raise CliError(
            "Hadoop encryption zone key management is not available without Active Directory authentication."
        )
    elif client.profile.active_context.security_strategy == SecurityStrategy.ad:
        client.login(endpoint)
    else:
        raise CliError(
            "Invalid security strategy.  Use azdata login to choose supported strategy."
        )

    return


def list_hdfskey(client):
    try:
        _login(client)
        result = client.list_hdfskey()
        return result
    except Exception as ex:
        raise_command_error(ex)


def create_hdfskey(client, name, size=256):
    try:
        _login(client)
        result = client.create_hdfskey(name, size)
        return result
    except Exception as ex:
        raise_command_error(ex)


def roll_hdfskey(client, name):
    try:
        _login(client)
        result = client.roll_hdfskey(name)
        return result
    except Exception as ex:
        raise_command_error(ex)

def describe_hdfskey(client, name):
    try:
        _login(client)
        status_code, response = client.describe_key(name)
        if status_code == 200:
            return convert_key_description_to_list(response)
        else:
            return response
    except Exception as ex:
        raise_command_error(ex)

def convert_key_description_to_list(key_description):
    """
    Converts the key description to a table format.
    """
    dict_key_key_version = "keyVersion"
    dict_key_main_key_version = "mainKeyVersion"
    row = {}
    row["name"] = key_description["name"]
    row["length"] = key_description["length"]
    row["created"] = key_description["created"]
    row["cipher"] = key_description["cipher"]
    row[dict_key_key_version] = key_description["attributes"][dict_key_key_version] \
        if dict_key_key_version in key_description["attributes"] else None
    row[dict_key_main_key_version] = key_description["attributes"][dict_key_main_key_version] \
        if dict_key_main_key_version in key_description["attributes"] else None
    return row