# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

from azdata.cli.commands.notebook.exceptions import NotebookEnvironmentError
from azdata.cli.core.configuration import Configuration
from azdata.cli.core.logging import get_logger
from azdata.cli.core.text import Text
from IPython.utils.tempdir import TemporaryDirectory
from jupyter_client.kernelspec import KernelSpecManager, NoSuchKernel

import os
import sys
import venv
import json

__all__ = ["NotebookEnvironmentFactory", "NotebookEnvironment"]

logger = get_logger(__name__)


# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #


class NotebookEnvironmentFactory(object):
    @staticmethod
    def create(env_name="base", *args, **kwargs):
        """
        Builds environments for notebooks to run under by environment name.

        :param env_name: The name of the environments to run a notebook in.
        :param display_name:
        :param system_site_packages: If True, the system (global) site-packages
                                     dir is available to created environments.
        :param clear: If True, delete the contents of the environment directory
          if
                      it already exists, before environment creation.
        :param symlinks: If True, attempt to symlink rather than copy files into
                         virtual environment.
        :param upgrade: If True, upgrade an existing virtual environment.
        :param with_pip: If True, ensure pip is installed in the virtual
                         environment.
        :return: The notebook environment for this `env_name`.
        """

        env_dir = None
        notebook_dir = NotebookEnvironment.NOTEBOOK_DIR

        if not os.path.exists(notebook_dir):
            os.mkdir(notebook_dir)

        for d in os.listdir(notebook_dir):
            if d == env_name:
                env_dir = os.path.join(notebook_dir, env_name)

        if not env_dir:
            if env_name == NotebookEnvironment.BASE_ENV:
                msg = "Creating new base environment for notebooks to run in..."
            else:
                msg = (
                    "No environment found for '{env}', defaulting to the "
                    "'{base}' environment.".format(
                        env=env_name,
                        base=NotebookEnvironment.BASE_ENV_DISPLAY_NAME,
                    )
                )

            Text.warning(msg)

            env_name = NotebookEnvironment.BASE_ENV
            env_dir = NotebookEnvironment.BASE_ENV_DIR

        env = NotebookEnvironment(env_name, *args, **kwargs)
        env.create(env_dir)

        return env


# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #


class NotebookEnvironment(venv.EnvBuilder):
    """
    This class exists to allow virtual environment creation to be
    customized. The constructor parameters determine the builder's
    behaviour when called upon to create a virtual environment.

    By default, the builder makes the system (global) site-packages dir
    *un*available to the created environment.

    If invoked using the Python -m option, the default is to use copying
    on Windows platforms but symlinks elsewhere. If instantiated some
    other way, the default is to *not* use symlinks.

    :param name:
    :param display_name:
    :param system_site_packages: If True, the system (global) site-packages
                                 dir is available to created environments.
    :param clear: If True, delete the contents of the environment directory if
                  it already exists, before environment creation.
    :param symlinks: If True, attempt to symlink rather than copy files into
                     virtual environment.
    :param upgrade: If True, upgrade an existing virtual environment.
    :param with_pip: If True, ensure pip is installed in the virtual
                     environment.
    :param pr   ompt: Alternative terminal prefix for the environment.
    """

    NOTEBOOK_DIR = os.environ.get("AZDATA_NOTEBOOK_ENV") if \
        os.environ.get("AZDATA_NOTEBOOK_ENV", None) else \
        os.path.join(Configuration().CLI_CONFIG_DIR, "notebooks")
    """
    The base location where to store notebook Python environments.
    """

    BASE_ENV = "base"
    BASE_ENV_DISPLAY_NAME = "azdata-cli (base)"
    BASE_ENV_DIR = os.path.join(NOTEBOOK_DIR, BASE_ENV)

    def __init__(self, name, *args, **kwargs):
        self._env_dir = None
        self._name = name
        self.kernel_name = None
        self._display_name = kwargs.pop("display_name", self.name)

        if self.name == self.BASE_ENV:
            self._display_name = self.BASE_ENV_DISPLAY_NAME

        # default to always include pip in env if not specified
        kwargs["with_pip"] = kwargs.get("with_pip", True)

        super().__init__(*args, **kwargs)

    @property
    def name(self):
        """
        Gets the name of this notebook environment.
        """
        return self._name

    @property
    def env_dir(self):
        """
        Gets the path of this notebook environment directory.
        """
        return self._env_dir

    
    @property
    def kernel_spec(self):
        """
        Gets the `KernelSpec` instance for the given `kernel_name`.
        :return: :class:`KernelSpec` instance for the given `kernel_name`.
        """
        if not self.kernel_name:
            return None
        try:
            return KernelSpecManager().get_kernel_spec(self.kernel_name)
        except NoSuchKernel:
            return None

    def create(self, env_dir):
        """
        Create a virtual environment in a directory if not already present.

        :param env_dir: The target directory to create an environment in.
        """

        logger.debug("Scanning for existing environment '%s'...", self.name)

        self._env_dir = env_dir

        if not os.path.exists(env_dir):
            try:
                os.makedirs(env_dir)
                super().create(env_dir)
            except Exception as e:
                logger.error(e)
        else:
            self.register_kernel_spec(kernel=None)

    def set_kernel_name(self, kernel_name):
        self.kernel_name = kernel_name

    def register_kernel_spec(self, kernel=None):
        """
        Register a new Kernel spec associated with this virtual environment
        context. This involves creating a new default `kernel.json` or using an
        existing one if a `kernel` is provided and installing it into the
        `KernelSpecManager`. The Kernel belonging to this Kernel spec will be
        guaranteed to run the Python interpreter that belongs to this
        environment.

        :param kernel: The optional kernel name to register and use.
        """
        p = os.path
        context = self.ensure_directories(self.env_dir)
        specs = KernelSpecManager().find_kernel_specs()
        env_dir = context.env_dir
        bin_path = context.bin_path
        python = (
            context.env_exe
            if sys.platform == "win32"
            else p.join(bin_path, "python")
        )

        kernel_json = {
            "argv": [
                python,
                "-m",
                "ipykernel_launcher",
                "-f",
                "{connection_file}",
            ],
            "display_name": self._display_name,
            "language": "python",
        }

        logger.debug(context)

        # Assert the env has a proper director structure: interpreter, lib
        if not p.exists(python):
            msg = (
                "Environment '{name}' is incorrectly configured at "
                "location:\n'{dir}'".format(name=self.name, dir=env_dir)
            )
            raise NotebookEnvironmentError(msg)

        # If kernel name provided attempt to use that kernel.json w/this-env
        if kernel:
            kpath = p.join(specs.get(kernel, ""), "kernel.json")
            if kpath and p.exists(kpath):
                logger.debug("Cloning kernel.json at '%s'", kpath)
                with open(kpath, "r", encoding="utf-8") as f:
                    kernel_json = json.load(f)
                    kernel_json["argv"][0] = python
            else:
                kernel = None

        # env exists, now make sure a kernel-spec is registered with it
        if not kernel and self.kernel_spec:
            logger.debug(
                "Environment '%s' exists, using this environment at location: "
                "'%s'",
                self.name,
                env_dir,
            )

            return

        logger.debug(
            "Registering a new kernel spec associated with environment '%s', "
            "from location '%s'\n%s",
            self.name,
            env_dir,
            kernel_json,
        )

        # -- Base Python3
        with TemporaryDirectory() as td:
            os.chmod(td, 0o755)
            with open(os.path.join(td, "kernel.json"), "w") as f:
                json.dump(kernel_json, f, sort_keys=True, indent=3)

            ksm = KernelSpecManager()
            ksm.install_kernel_spec(
                td, kernel_name=self.kernel_name, user=True
            )

        # -- If not found install default `pysparkkernel` --
        if not specs.get("pysparkkernel"):
            pyspark_source_dir = p.join(
                self.site_packages(env_dir),
                "sparkmagic",
                "kernels",
                "pysparkkernel"
            )

            logger.debug("No existing 'pysparkkernel' found, installing"
                         " default 'pysparkkernel' from source location: %s",
                         pyspark_source_dir)

            KernelSpecManager().install_kernel_spec(
                pyspark_source_dir, user=True
            )


        # -- If not found install default `sparkrkernel` --
        if not specs.get("sparkrkernel"):
            sparkr_source_dir = p.join(
                self.site_packages(env_dir),
                "sparkmagic",
                "kernels",
                "sparkrkernel"
            )

            logger.debug("No existing 'sparkrkernel' found, installing"
                         " default 'sparkrkernel' from source location: %s",
                         sparkr_source_dir)

            KernelSpecManager().install_kernel_spec(
                sparkr_source_dir, user=True
            )


        # -- If not found install default `sparkkernel` --
        if not specs.get("sparkkernel"):
            spark_source_dir = p.join(
                self.site_packages(env_dir),
                "sparkmagic",
                "kernels",
                "sparkkernel"
            )

            logger.debug(
                "No existing 'sparkkernel' found, installing default "
                "'sparkkernel' from source location: %s", spark_source_dir)

            KernelSpecManager().install_kernel_spec(spark_source_dir, user=True)

    @staticmethod
    def site_packages(env_dir):
        if sys.platform == "win32":
            libpath = os.path.join(env_dir, "Lib", "site-packages")
        else:
            libpath = os.path.join(
                env_dir,
                "lib",
                "python%d.%d" % sys.version_info[:2],
                "site-packages",
            )
        return libpath

    def post_setup(self, context):
        """
        Set up any packages which need to be pre-installed into the
        virtual environment being created.

        :param context: The information for the virtual environment
                        creation request being processed.
        """
        import zipfile

        cwd = os.path.abspath(os.path.dirname(__file__))
        env_zip = os.path.join(cwd, ".env", "env.zip")
        env_dir = context.env_dir
        libpath = self.site_packages(env_dir)

        logger.debug("========================================================")
        logger.debug(
            "No environment found for '%s', a new (base) "
            "environment was created: '%s'",
            self.name,
            env_dir,
        )

        logger.debug("Environment site-packages location '%s'", libpath)

        if not os.path.exists(env_zip):
            msg = "Installing base packages, this can take several minutes..."
            logger.debug("No base environment packages found \n'%s'", env_zip)
            logger.debug(msg)

            Text.warning(msg)

            self.__seed__(libpath)
        else:
            logger.debug("Unzipping base environment packages from '%s'",
                         env_zip)

            with zipfile.ZipFile(env_zip, "r") as z:
                z.extractall(libpath)

        self.register_kernel_spec()

        logger.debug("========================================================")

    def __seed__(self, target):
        """
        Seed env with base packages.
        """
        # pylint: disable=protected-access
        from pip._internal.cli.main import main as pip_install
        from azdata.cli.commands.notebook.__base_packages__ import packages

        try:
            cmd = [
                "install",
                "--target={}".format(target),
                "--quiet",
                "--no-warn-conflicts",
                "--disable-pip-version-check",
                "--upgrade",
                "pip==20.2.4"
            ]

            for pkg in packages:
                Text.warning("Installing {pkg}".format(pkg=pkg))
                cmd.append(pkg)
                pip_install(cmd)
                cmd.pop()

        except (SystemExit, Exception) as err:
            print(err)
