# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

from nbclient import NotebookClient as NBClient
from nbclient.exceptions import CellExecutionError, CellTimeoutError
from nbformat import from_dict
from nbconvert import NotebookExporter, HTMLExporter
from azdata.cli.core.logging import get_logger
from azdata.cli.commands.notebook.util import (
    CellFormatter,
    WorkingDirectoryContext,
    decode_dictionary,
)

import os
import json
import codecs
import tempfile

try:
    import sparkmagic.utils.configuration as conf
    conf.override("cleanup_all_sessions_on_exit", True)
    # Note for future: we can set "ignore_ssl_errors" to True, 
    # but we would also need to let the users know about it 
    # and possibly allow to configure it on-demand:
    # conf.override("ignore_ssl_errors", True)
except Exception:
    pass

__all__ = ["NotebookClient"]

logger = get_logger(__name__)


class NotebookClient(object):
    DEFAULT_KERNEL_NAME = "python3"
    """
    Defines the default kernel name which represents that native py interpreter.
    """

    def __init__(self, path, env):
        self._path = path
        self._env = env

        with open(self._path, "r", encoding="utf-8") as f:
            self._notebook = json.load(f)
            self.kernel_spec = self.extract_kernel_spec(self._notebook)
            self.kernel_name = self.extract_kernel_name(self.kernel_spec)

            # If this notebook has a kernel spec provided in its top-level metadata,
            # register it with this environment otherwise the default is already
            # registered.
            #
            if self.kernel_spec:
                self._env.register_kernel_spec(kernel=self.kernel_name)
                self._env.set_kernel_name(kernel_name=self.kernel_name)

    def extract_kernel_name(self, kernel_spec):
        default = self.DEFAULT_KERNEL_NAME
        return kernel_spec["name"] if kernel_spec else default

    def extract_kernel_spec(self, notebook):
        kernel_spec = None

        if (
            "metadata" in notebook
            and "kernelspec" in notebook["metadata"]
            and "name" in notebook["metadata"]["kernelspec"]
        ):
            kernel_spec = notebook["metadata"]["kernelspec"]

        return kernel_spec

    def interact(self, clear=None):
        from azdata.cli.commands.notebook.interactive import InteractiveNotebook

        InteractiveNotebook.launch_interactive_console_notebook(
            self._env, self._path, clear
        )

    def render(self, continue_on_error):
        output = ""
        i = 1

        for cell in self._notebook["cells"]:
            cell_output, found_error = CellFormatter.render_cell(
                cell, self.kernel_name, i
            )
            output += cell_output
            i += 1
            if found_error and not continue_on_error:
                break

        return output

    def run(
        self,
        client,
        output_path=None,
        output_html=None,
        arguments=None,
        timeout=60,
        suppress_generated_cells=False,
    ):
        def _write_to_file(filename, body):
            with codecs.open(filename, mode="w", encoding="utf-8") as file:
                file.write(body)

        def _notebook_to_json(notebook_node):
            return NotebookExporter().from_notebook_node(notebook_node)

        def _notebook_to_html(notebook_node):
            exporter = HTMLExporter()
            exporter.template_file = "full"
            return exporter.from_notebook_node(notebook_node)

        execution_exception = None
        html_notebook_output_file = None
        rm_index = -1
        max_retry = 3
        cf = CellFormatter

        # Setup the output directory and file names based on the input
        # parameters
        #
        notebook_path = os.path.abspath(self._path)
        notebook_directory, notebook_filename = os.path.split(notebook_path)
        output_notebook_filename = "output-" + notebook_filename

        if output_path == None:
            working_folder = tempfile.mkdtemp()
            notebook_output_file = os.path.join(
                working_folder, output_notebook_filename
            )
        else:
            working_folder = output_path
            notebook_output_file = os.path.join(
                output_path, output_notebook_filename
            )

        resources = {"metadata": {"path": working_folder}}

        if output_html:
            file_prefix, file_ext = os.path.splitext(output_notebook_filename)
            html_notebook_output_file = os.path.join(
                working_folder, "{0}.{1}".format(file_prefix, "html")
            )

        #with WorkingDirectoryContext(working_folder):
        with WorkingDirectoryContext(working_folder) and open(
                    notebook_path, "r", encoding="utf-8"
        ) as f:

            self._notebook = json.load(f)

            # Decode the file into nested dictionary data structure
            #
            notebook = self._notebook

            # Load the name of the Jupyter kernel configured in the notebook.
            # If the notebook doesn't have a kernel configured then Python3 is
            # used as the default.
            #
            kernel_name = self.kernel_name
            is_spark = self.is_spark()

            logger.debug("Spark kernel: %s", is_spark)

            # Process passed in arguments
            #
            # NOTE: for now this only supports Python based kernels.
            #
            # if kernel_name == "python3" or kernel_name == "pysparkkernel" or "sparkrkernel":
            if kernel_name == self.DEFAULT_KERNEL_NAME or is_spark:
                if arguments is not None:
                    # If the user has passed in arguments then see if the
                    # notebook has a cell with a 'parameters' tag. If it does
                    # generate a new cell after the parameters cell with the
                    # passed in parameters.
                    #
                    arguments = decode_dictionary(arguments)
                    cf.generate_parameters_cell(
                        notebook, kernel_name, arguments
                    )

                # Generate kernel specific cells
                #
                rm_index = cf.generate_kernel_specific_cells(
                    client, notebook, is_spark
                )

            # Canonicalize the notebook json to what the client expects.
            #
            notebook = cf.canonicalize_code_cells(notebook)

            # Convert from json to NotebookNode
            #
            nbnode = from_dict(notebook)

            # Execute the notebook.  The output is a new NotebookNode containing
            # the results.
            #
            retry_count = 0
            done = False
            completed_notebook = None

            while not done:
                nbclient = NBClient(nbnode)
                nbclient.timeout = timeout
                nbclient.kernel_name = self._env.kernel_name
                nbclient.resources = resources

                try:
                    completed_notebook = nbclient.execute()
                    done = True
                except CellExecutionError as cee:
                    execution_exception = cee
                    completed_notebook = nbclient.nb
                    done = True
                except CellTimeoutError as cte:
                    execution_exception = cte
                    completed_notebook = nbclient.nb
                    done = True
                except RuntimeError as rte:
                    # The kernel died.  Try again
                    if retry_count < max_retry:
                        done = False
                        retry_count += 1
                    else:
                        execution_exception = rte
                        if nbclient and hasattr(nbclient, "nb"):
                            completed_notebook = nbclient.nb
                        done = True
                finally:
                    del nbclient

            if rm_index >= 0:
                completed_notebook = cf.remove_generated_cell_by_index(
                    completed_notebook, rm_index
                )

            # If there were generated cells remove them
            #
            if suppress_generated_cells:
                completed_notebook = cf.remove_generated_cells(
                    completed_notebook
                )

            # Convert output
            # TODO: handle resources correctly.  Right now they aren't saved to
            # the output directory
            #
            output_notebook, resources = _notebook_to_json(completed_notebook)

            # If specified convert the notebook to HTML
            #
            if output_html:
                (
                    html_notebook,
                    html_resources,
                ) = _notebook_to_html(completed_notebook)

            # Write the output files after input file is closed to support
            # overwrite
            #
            _write_to_file(notebook_output_file, output_notebook)
            if output_html:
                _write_to_file(html_notebook_output_file, html_notebook)

        return (
            notebook_output_file,
            html_notebook_output_file,
            execution_exception,
        )

    def is_spark(self):
        kernel_spec = self._env.kernel_spec.to_dict()
        argv = kernel_spec.get("argv")

        logger.debug("kernel: %s", kernel_spec)

        if argv and len(argv) >= 2:
            mod = argv[2]  # get module
            found_by_name = self.kernel_name in [
                "pysparkkernel",
                "pyspark3kernel",
                "sparkrkernel",
            ]

            return found_by_name or mod in [
                "sparkmagic.kernels.pysparkkernel.pysparkkernel",
                "sparkmagic.kernels.sparkrkernel.sparkrkernel",
                "sparkmagic.kernels.sparkkernel.sparkkernel",
            ]
        else:
            return False
