# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

from pygments import highlight
from pygments.lexers import PythonLexer, MarkdownLexer, HtmlLexer
from pygments.formatters import (
    TerminalFormatter,
    Terminal256Formatter,
    HtmlFormatter,
)
from azdata.cli.commands.notebook import _
from azdata.cli.core.exceptions import CliError
from azdata.cli.core.profile import CliProfile
from azdata.cli.core.enums import SecurityStrategy
from azdata.cli.core.output import OutputStream
from azdata.cli.core.prompt import prompt_pass
from knack.prompting import NoTTYException

import os
import mdv
import json

# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #


class CellFormatter(object):
    @staticmethod
    def convert_list_to_string(cell_content):
        # If the input type is a list merge them together since the formatter
        # takes a multiline string.
        if type(cell_content) is list:
            # If the array has embedded new lines then concatenate without
            # adding new lines.
            if "\n" in cell_content[0]:
                cell_content = "".join(cell_content)
            else:
                cell_content = "\n".join(cell_content)
        return cell_content

    @staticmethod
    def _render_markdown_cell(cell_content):
        cell_content = CellFormatter.convert_list_to_string(cell_content)
        mdv.term_columns = 60
        output = mdv.main(cell_content, theme="pastel", c_theme="pastel")
        return output

    @staticmethod
    def _render_python_cell(cell_content):
        cell_content = CellFormatter.convert_list_to_string(cell_content)
        output = highlight(cell_content, PythonLexer(), Terminal256Formatter())
        return output

    @staticmethod
    def _render_output_cell(cell_content):
        cell_content = CellFormatter.convert_list_to_string(cell_content)

        # Note that there might be other types of output content and right now
        # it's not getting formatted.
        output = "\n[Output]:\n"
        output += cell_content
        return output

    @staticmethod
    def _render_cell_boundary(cell_type, cell_number):
        return _(
            "\n[{0: <8}] CELL [{1: >4}] ---------------------------------------------------------------------------------------\n".format(
                cell_type, cell_number
            )
        )

    @staticmethod
    def render_cell(cell, kernel_type, cell_number):
        found_error = False
        output = CellFormatter._render_cell_boundary(
            cell["cell_type"], cell_number
        )

        if cell["cell_type"] == "markdown":
            output += CellFormatter._render_markdown_cell(cell["source"])
        elif cell["cell_type"] == "code":
            if kernel_type.startswith("python"):
                output += CellFormatter._render_python_cell(cell["source"])
            else:
                # If the cell type isn't known add the content without syntax
                # highlighting
                output += cell["source"]

            if len(cell["outputs"]) > 0:
                for output_cell in cell["outputs"]:
                    if "text" in output_cell:
                        output += CellFormatter._render_output_cell(
                            output_cell["text"]
                        )
                    if "traceback" in output_cell:
                        output += CellFormatter._render_output_cell(
                            output_cell["traceback"]
                        )
                    if output_cell["output_type"] == "error":
                        found_error = True

        return output, found_error

    ########################################################################
    @staticmethod
    def remove_generated_cells(notebook):
        generated_cells = CellFormatter.cells_with_tag(notebook, "generated")
        if len(generated_cells) > 0:
            generated_cells.sort(reverse=True)
            for index in generated_cells:
                CellFormatter.remove_generated_cell_by_index(notebook, index)

        return notebook

    @staticmethod
    def remove_generated_cell_by_index(notebook, index):
        if index >= 0:
            notebook["cells"].remove(notebook["cells"][index])

        return notebook

    @staticmethod
    def canonicalize_code_cells(notebook_json):
        """
        The Notebook executor expects code source to be in a string. JSON
        formatters use multi-line to make reading the code easier.  This method
        converts the source lists to strings.
        """
        for cell in notebook_json["cells"]:
            # If the input type is a list merge them together since the
            # formatter takes a multiline string.
            #
            if "source" in cell:
                source = cell["source"]
                if type(source) is list:
                    # If the array has embedded new lines then concatenate
                    # without adding new lines.
                    #
                    if len(source) > 0 and "\n" in source[0]:
                        cell["source"] = "".join(source)
                    else:
                        cell["source"] = "\n".join(source)

        return notebook_json

    @staticmethod
    def cells_with_tag(notebook, tag):
        """
        This is used to find cells with a particular tag.  Examples would be to
        find where parameters should be inserted or to find all the generated
        cells in a notebook.
        """
        tagged_cell_indexes = [
            notebook["cells"].index(x)
            for x in notebook["cells"]
            if x["cell_type"] == "code"
            and "tags" in x["metadata"]
            and tag in x["metadata"]["tags"]
        ]
        return tagged_cell_indexes

    @staticmethod
    def generate_cell(cell_body):
        # NOTE: Currently only python cells can be generated
        #
        cell = dict(
            cell_type="code", execution_count=None, outputs=[], source=[]
        )

        cell["metadata"] = {"tags": ["generated"]}

        cell["source"].append(
            "#----------------------------------------------------------------------------------------\n"
            "# Generated Cell\n"
            "#\n"
            "# This cell contains generated code.  These generated blocks can "
            "be created for several\n"
            "# reasons.  Examples include command line parameter injection or "
            "kernel specific bootstrap\n"
            "# code.  The generated cells should not contain sensitive data.  "
            "To pass sensitive data\n"
            "# to a notebook use an environment variable or use secure "
            "storage.\n"
            "#----------------------------------------------------------------------------------------\n"
            "#\n"
        )
        cell["source"].append("{0}".format(cell_body))

        cell["source"] = CellFormatter.convert_list_to_string(cell["source"])

        return cell

    @staticmethod
    def generate_parameters_cell(notebook, kernel_spec, arguments):
        msg = ""
        if arguments is not None and kernel_spec == "python3":
            # If the user has passed in arguments then see if the notebook has
            # a cell with a 'parameters' tag. If it does generate a new cell
            # after the parameters cell with the passed in parameters.
            #
            # NOTE for now this only supports Python cells.
            #
            cells = CellFormatter.cells_with_tag(notebook, "parameters")
            if len(cells) > 0:
                cell_body = ""
                for key in sorted(arguments):
                    cell_body += '{0} = "{1}"\n'.format(key, arguments[key])

                gen_cell = CellFormatter.generate_cell(
                    CellFormatter.convert_list_to_string(cell_body)
                )
                notebook["cells"].insert(cells[0] + 1, gen_cell)
            else:
                # None of the cells were marked for parameters.
                #
                msg += (
                    "Warning: Parameters were passed in but no cells in the "
                    "target notebook were marked with parameters tag.  "
                    "Parameters not inserted.\n"
                )
        return msg

    @staticmethod
    def generate_kernel_specific_cells(client, notebook, is_spark):
        rm_index = -1

        if is_spark:
            profile = CliProfile()
            rm_index = 0

            # Need to generate the _do_not_call_change_endpoint in order to
            # setup the kernel connection to Spark. This uses the Spark command
            # module code to get the connection properties.
            endpoint = get_livy_endpoint(client)

            if (
                profile.active_context.security_strategy
                == SecurityStrategy.basic
            ):
                username = get_knox_username()
                password = get_knox_password()
                spark_magic = (
                    "%_do_not_call_change_endpoint "
                    "--username={0} "
                    "--password={1} "
                    "--server={2} "
                    "--auth=Basic_Access".format(username, password, endpoint)
                )
            elif (
                profile.active_context.security_strategy == SecurityStrategy.ad
            ):
                spark_magic = (
                    "%_do_not_call_change_endpoint "
                    "--server={0} "
                    "--auth=Kerberos".format(endpoint)
                )
            else:
                raise CliError(
                    "Invalid security strategy.  Use azdata login to choose "
                    "supported strategy."
                )

            generated_cell = CellFormatter.generate_cell("")

            # Replace the source of the generated cell.  The only thing in the
            # cell can be the spark magic. By having the cell comment the client
            # tries to send the source to the server and can't connect.
            #
            generated_cell["source"] = spark_magic

            # Execute the cell instead of inserting it into the notebook.  This
            # is because this generated cell can contain passwords and that
            # shouldn't be persisted to disk.
            #
            notebook["cells"].insert(0, generated_cell)

        return rm_index


# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #


class WorkingDirectoryContext(object):
    def __init__(self, working_path):
        self.working_path = os.path.expanduser(working_path)

    def __enter__(self):
        self.saved_path = os.getcwd()
        os.chdir(self.working_path)

    def __exit__(self, etype, value, traceback):
        os.chdir(self.saved_path)


# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #


def dump_buffer(buffer):
    output = ""
    length = len(buffer)
    for offset in range(0, length, 16):
        data = buffer[
            offset : offset + 16 if offset + 16 <= length else length - offset
        ]
        output += "{:08X} | {:47} | {}\n".format(
            offset,
            " ".join("{:02X}".format(i) for i in data),
            "".join(31 < i < 127 and chr(i) or "." for i in data),
        )
    return output


def diff_buffers(buffer1, buffer2, show_equal=True):
    output = ""
    length = max(len(buffer1), len(buffer2))
    for offset in range(0, length, 16):
        data_buffer1 = buffer1[
            offset : offset + 16 if offset + 16 <= length else length - offset
        ]
        data_buffer2 = buffer2[
            offset : offset + 16 if offset + 16 <= length else length - offset
        ]

        if data_buffer1 == data_buffer2:
            equal = True
        else:
            equal = False

        if equal == False or (show_equal == True and equal == True):
            output += "{:08X} | {:47} | {:16} {:4} {:47} | {:16}\n".format(
                offset,
                " ".join("{:02X}".format(i) for i in data_buffer1),
                "".join(31 < i < 127 and chr(i) or "." for i in data_buffer1),
                " == " if equal else " XX ",
                " ".join("{:02X}".format(i) for i in data_buffer2),
                "".join(31 < i < 127 and chr(i) or "." for i in data_buffer2),
            )

    return output


def get_livy_endpoint(client):
    """
    Call the controller to get the webhdfs endpoint URL
    """
    try:
        endpoint = client.apis.controller.get_endpoints("livy")["endpoint"]

        # The endpoint must end with '/' in order for the APIs to succeed
        if not endpoint[len(endpoint) - 1] == "/":
            endpoint = endpoint + "/"
    except Exception as e:
        raise CliError(e)

    return endpoint


def get_knox_username():
    if not os.getenv("AZDATA_USERNAME"):
        try:
            stdout = OutputStream().stdout.write
            stdout(
                "Username is needed.  It can be set for the session by"
                " setting the environment variable AZDATA_USERNAME."
            )
            username = prompt_pass("Enter username: ")
        except NoTTYException:
            raise NoTTYException(
                "Missing AZDATA_USERNAME environment variable "
                "in non-interactive environment."
            )
    else:
        username = os.environ["AZDATA_USERNAME"]

    return username



def get_knox_password():
    # Note for now if the environment variable isn't set prompt for the
    # password.

    if not os.getenv("AZDATA_PASSWORD"):
        try:
            stdout = OutputStream().stdout.write
            stdout(
                "The Knox password is needed.  It can be set for the session by"
                " setting the environment variable AZDATA_PASSWORD."
            )
            password = prompt_pass("Enter Knox Password: ")
        except NoTTYException:
            raise NoTTYException(
                "Missing AZDATA_PASSWORD environment variable "
                "in non-interactive environment."
            )
    else:
        password = os.environ["AZDATA_PASSWORD"]

    return password


def decode_dictionary(value):
    """
    Some parameters need to be dictionary of name value pairs.  For example
    configuration dictionary. This function checks to see if the content is
    a json dictionary and if so decodes it.

    Example:
    { "name": "value", "another_name": "another_value" }
    """
    if value.strip().startswith("{"):
        result = json.loads(value)
    else:
        raise ValueError(
            "Value must be JSON encoded dictionary [{0}].".format(value)
        )

    return result
