# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

from azdata.cli.core.enums import ContextType
from azdata.cli.core.clients.cli_client import CliClient
from azdata.cli.core.profile import (
    CliProfile,
    select_authenticator
)


def beget(_):
    """Client factory for user's authentication profile."""
    return ProfileClientMixin()


class ProfileClientMixin(CliClient):
    """
    User profile client.
    """

    def __init__(self):
        super(ProfileClientMixin, self).__init__(
            check_auth=False, check_eula=False)

        self.active_context = self.profile.active_context

    def set_profile(
            self,
            accept_eula,
            endpoint,
            username,
            security_strategy,
            token,
            context_type):

        # -- persist --
        CliProfile(
            accept_eula=accept_eula,
            endpoint=endpoint,
            username=username,
            security_strategy=security_strategy,
            auth_token=token,
            context_type=context_type
        )

    def login(self, args):
        authenticator = select_authenticator(args)
        token = authenticator.authenticate()

        # -- persist token and endpoint into the profile --
        self.set_profile(
            args.get('accept_eula', 'no'),
            authenticator.endpoint,
            authenticator.username,
            args.get('auth'),
            token,
            ContextType.bdc.name if token.context_type is None else token.context_type
        )

        return token.namespace

    def logout(self):
        try:
            profile = CliProfile()
            if profile.active_context and profile.active_context.namespace:
                self.delete_context(
                    profile.active_context.namespace,
                    profile.active_context.username,
                    profile.active_context.endpoint
                )
        except Exception as e:
            raise e

    def endpoint(self):
        return self.active_context.endpoint

    def accept_eula(self):
        return self.profile.accept_eula

    def namespace(self):
        return self.active_context.namespace

    def username(self):
        return self.active_context.username

    def list_context(self, active=False):
        try:
            if not active:
                contexts = CliProfile().contexts
                if contexts:
                    for key, value in contexts.items():
                        contexts[key]['accessToken'] = '<PRIVATE>'
                    return contexts
                else:
                    return None
            else:
                if CliProfile().active_context.namespace:
                    context = CliProfile().active_context.serialize()
                    context['accessToken'] = '<PRIVATE>'
                    return context
                else:
                    return None
        except Exception as e:
            raise ValueError('Failed to list contexts - {0}'.format(e))

    def set_context(self, namespace, username=None, endpoint=None):
        try:
            key = self.get_context_key(namespace, username, endpoint)
            profile = CliProfile()
            profile.active_context = profile.contexts[key]
        except Exception as e:
            raise ValueError('Failed to set context as {0} - {1}'.
                             format(namespace, e))

    def delete_context(self, namespace, username=None, endpoint=None):
        try:
            key = self.get_context_key(namespace, username, endpoint)
            profile = CliProfile()
            contexts = profile.contexts
            del contexts[key]

            if namespace == profile.active_context.namespace:
                profile.active_context = None
            if contexts:
                profile.contexts = contexts
            else:
                profile.contexts = None

        except Exception as e:
            raise ValueError('Failed to delete {0} - {1}'.format(namespace, e))

    def get_context_key(self, namespace, username, endpoint):
        full_namespace = all(arg is not None for arg in
                             [namespace, username, endpoint])
        hits = 0
        context_key = None
        profile = CliProfile()

        if profile.contexts:
            for key, context in profile.contexts.items():
                if namespace == context.get('namespace'):
                    hits += 1

                    if full_namespace:
                        if username == context.get('username') and \
                                        endpoint == context.get('endpoint'):
                            context_key = key
                    else: # namespace only
                        context_key = key
        else:
            raise ValueError('No contexts exist.  Please login to a '
                             'control plane.')

        if not context_key:
            if full_namespace:
                raise ValueError('Context does not exist with:'
                                 '\nnamespace "{0}" '
                                 '\nusername "{1}"'
                                 '\nendpoint "{2}"'.
                                 format(namespace, username, endpoint))
            else:
                raise ValueError('Context with namespace "{0}" does not exist.'.
                             format(namespace))

        if hits > 1 and not full_namespace:
            raise ValueError(
                'Multiple contexts exist with this namespace '
                '"{0}".\nPlease include all [--namespace, '
                '--username, --endpoint] for context uniqueness.'.
                    format(namespace)
            )

        return context_key
