# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------
from datetime import datetime as dt
import threading
import time

from urllib3.exceptions import NewConnectionError, MaxRetryError

from azdata.cli.core.clients.cli_client import CliClient
from azdata.cli.core.deploy import display
from azdata.cli.core.logging import get_logger
import azdata.cli.core.progress as progress_handler
import azdata.cli.core.util as core_util
import azdata.cli.core.deploy as util
import azdata.cli.core.kubernetes as kubernetes_util
import azdata.cli.core.upgrade as upgrade_util


CONNECTION_RETRY_ATTEMPTS = 12
RETRY_INTERVAL = 5
UPDATE_INTERVAL = (15 * 60) / RETRY_INTERVAL
logger = get_logger(__name__)

__all__ = ["beget"]


def beget(_):
    """Client factory"""
    return RotateClientMixin(check_auth=False, check_eula=False)


class RotateClientMixin(CliClient):
    def __init__(self, check_auth=False, check_eula=False):
        """
        :param check_auth: Set to `True` to have this command enforce
                           authentication before running.
        :param check_auth: Set to `True` to have this command enforce eula
                           acceptance before running.
        """

        self._timestamp = dt.fromtimestamp(dt.timestamp(dt.now()))

        super(RotateClientMixin, self).__init__(
            check_auth=check_auth, check_eula=check_eula
        )

    def rotate_bdc(self, cluster_name):
        """
        Rotate credentials in a cluster
        """
        self.cluster_name = cluster_name

        kubernetes_util.validate_namespace(self.cluster_name)

        # Make sure the namespace exists
        #
        if not util.retry(
            kubernetes_util.namespace_exists,
            self.cluster_name,
            retry_count=CONNECTION_RETRY_ATTEMPTS,
            retry_delay=RETRY_INTERVAL,
            retry_method="check if namespace exists",
            retry_on_exceptions=(NewConnectionError, MaxRetryError),
        ):
            raise Exception(
                "Rotating cluster credentials failed since the "
                "namespace %s does not exist." % self.cluster_name
            )

        if not self.apis.control_deploy.is_controller_pod_running(
            self.cluster_name
        ):
            raise Exception(
                "Rotating cluster credentials failed since the "
                "controller is not deployed into namespace %s."
                % self.cluster_name
            )

        display("Upgrading Control Plane.")

        self.apis.controller.rotate_credentials()

        # All control upgrade monitoring will be replaced by calls to
        # the monitoring pod. Until then we will manually check the
        # upgrade config to see the status of the controller upgrade, and
        # the control plane state machine for the rest of the upgrade.
        #
        i = 0
        while upgrade_util.is_controller_upgrading(self.cluster_name):
            if i != 0 and i % UPDATE_INTERVAL == 0:
                logger.warning(
                    "Controller is still rotating credentials after %d "
                    "minutes.",
                    core_util.minutes_elapsed(i, RETRY_INTERVAL),
                )

            time.sleep(RETRY_INTERVAL)
            i = i + 1

        if upgrade_util.check_controller_upgrade_failed(self.cluster_name):
            raise Exception(
                "Credential rotation failed. "
                "Failed to rotate AD credentials."
            )

        # The controller and controller db have been upgraded.
        # Now we need to wait for the rest service to be reenabled
        #
        i = 0
        while not self.apis.controller.is_ready():
            time.sleep(RETRY_INTERVAL)
            i = i + 1

        control_plane_state = str(self.apis.controller.get_control_status(
            control_plane_state=True
        )).lower()

        stop = threading.Event()
        status_thread = threading.Thread(
            target=self.apis.control_deploy.get_controller_logs,
            args=[stop, False],
        )
        status_thread.setDaemon(True)
        status_thread.start()

        i = 0
        while upgrade_util.is_control_plane_upgrading(self.cluster_name):

            if control_plane_state == "error":
                stop.set()
                status_thread.join()
                progress_handler.notify_control_plane_upgrade_error(
                    self.cluster_name, control_plane_state
                )
                raise Exception(
                    "Control plane credential rotation failed. "
                    "Cluster State: %s." % control_plane_state
                )
            if control_plane_state == "upgradepaused":
                # Display warning message for configuring SQL
                # server settings if required.
                #
                progress_handler.notify_control_plane_upgrade_paused()
                return

            if i != 0 and i % UPDATE_INTERVAL == 0:
                logger.warning(
                    "Control plane is not ready after %d " "minutes. ",
                    core_util.minutes_elapsed(i, RETRY_INTERVAL),
                )

            if control_plane_state is not None:
                logger.debug(
                    "Waiting for Control plane to be in Ready "
                    "state. Current state : %s",
                    control_plane_state,
                )

            time.sleep(RETRY_INTERVAL)
            control_plane_state = str(self.apis.controller.get_control_status(
                control_plane_state=True)).lower()
            i = i + 1

        cluster_state = self.apis.controller.get_cluster_state()

        # It can take a second or two for the state machine to move back
        # into the upgrading state. If the upgrade is still old, retry.
        # We don't pause an upgrade until it has been updating for at
        # least ten minutes, so we shouldn't get stuck in this loop.
        #
        while cluster_state == "upgradepaused":
            time.sleep(RETRY_INTERVAL)

            cluster_state = self.apis.controller.get_cluster_state()

        i = 0
        while upgrade_util.are_cluster_credentials_rotating(self.cluster_name):
            if cluster_state == "error":
                stop.set()
                status_thread.join()
                progress_handler.notify_bdc_upgrade_error(
                    self.cluster_name, cluster_state
                )
                raise Exception(
                    "Cluster '{0}' credential rotation failed. "
                    "Cluster State: {1}".format(
                        self.cluster_name, cluster_state
                    )
                )
            if cluster_state == "upgradepaused":
                # Display warning message for configuring SQL server
                # settings if required.
                #
                progress_handler.notify_bdc_upgrade_paused()
                return

            if i != 0 and i % UPDATE_INTERVAL == 0:
                logger.warning(
                    "Cluster '%s' is not ready after" " %d minutes.",
                    self.cluster_name,
                    core_util.minutes_elapsed(i, RETRY_INTERVAL),
                )

            if cluster_state is not None:
                logger.debug(
                    "Waiting for cluster '%s' to be in Ready "
                    "state. Current state : %s",
                    cluster_name,
                    cluster_state,
                )

            time.sleep(RETRY_INTERVAL)
            cluster_state = self.apis.controller.get_cluster_state()
            i = i + 1

        progress_handler.notify_bdc_upgrade_success()
