# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

"""
Entry or launch point for the CLI.

Handles creating and launching the CLI to handle a user commands.
"""

from azdata.cli.core.commands import CliCommandsLoader
from azdata.cli.core.logging import CliLogging
from azdata.cli.core.configuration import Configuration
from azdata.cli.core.output import OutputProducer
from azdata.cli.core.util import env_to_section_option

import knack
import knack.help
import knack.config
import azdata.cli.core.help
import os
import sys

WELCOME_MESSAGE = r"""
    ___ _____   ____  ___  _________ 
   /   /__  /  / __ \/   |/_  __/   |
  / /| | / /  / / / / /| | / / / /| |
 / ___ |/ /__/ /_/ / ___ |/ / / ___ |
/_/  |_/____/_____/_/  |_/_/ /_/  |_|

Welcome to the {cli_name} CLI.

Use `{cli_name} --version` to display the current version.
Here are the base commands:
""".format(cli_name=Configuration().CLI_NAME)


def cli():
    """
    Entry point for the CLI.

    Configures and invokes the CLI with arguments passed during the time the
    python session was initiated.
    """
    config = Configuration()

    class CliHelp(knack.help.CLIHelp):
        def __init__(self, cli_ctx=None):
            super(CliHelp, self).__init__(
                cli_ctx=cli_ctx,
                privacy_statement='The privacy statement can be viewed at:\n'
                                  'https://go.microsoft.com/fwlink/?LinkId='
                                  '2002534',
                welcome_message=WELCOME_MESSAGE
            )

    class CliConfig(knack.config.CLIConfig):
        def __init__(self, config_dir=config.CLI_CONFIG_DIR,
                     config_env_var_prefix=config.CLI_NAME.upper()):
            super(CliConfig, self).__init__(
                config_dir=config_dir,
                config_env_var_prefix=config_env_var_prefix
            )

            envs = {}
            prefix = config_env_var_prefix

            # -- CLI Configuration ENV defaults --
            envs['{}_extension_dir'.format(prefix)] = None
            envs['{}_logging_log_dir'.format(prefix)] = None
            envs['{}_logging_enable_log_file'.format(prefix)] = 'Yes'
            envs['{}_http_verify_ssl'.format(prefix)] = 'No'
            envs['{}_http_ssl_ca_cert'.format(prefix)] = None
            envs['{}_http_cert_file'.format(prefix)] = None
            envs['{}_http_key_file'.format(prefix)] = None
            envs['{}_http_assert_hostname'.format(prefix)] = 'No'
            envs['{}_http_proxy_url'.format(prefix)] = None
            envs['{}_output_interactive'.format(prefix)] = 'Yes' if \
                sys.stdin.isatty() else 'No'

            for env in envs:
                section, option = env_to_section_option(env)
                value = self.get(section, option, fallback=envs[env])

                # if env was set use this, otherwise take the config or default
                # env > config > default
                if os.environ.get(env.upper()):
                    envs[env] = os.environ.get(env.upper(), value)

                # Apply and persisted config values or env values to the
                # runtime configuration
                if hasattr(config, option):
                    setattr(config, option, envs[env])

            self.set_values(envs)

        def set_values(self, items):
            """
            Persist all values to `config` file in one write.
            """
            from knack.config import configparser, get_config_parser
            cli_config = get_config_parser()
            cli_config.read(self.config_path)

            # -- include any special configuration items --
            if cli_config.has_section('azdatadev'):
                house = cli_config.get('azdatadev',
                                       'extension_house', fallback=None)
                house = house or config.extension_house  # else use default
                config.extension_house = house

            # -- save env-var values --
            for item in items:
                section, option = env_to_section_option(item)
                try:
                    cli_config.add_section(section)
                except configparser.DuplicateSectionError:
                    pass
                if items[item]:
                    cli_config.set(section, option, items[item])

            self._config_file_chain[0].set(cli_config)  # persist

    class GlobalContextArgument(object):
        def __init__(self, cli_ctx):
            from knack.events import (EVENT_PARSER_GLOBAL_CREATE,
                                      EVENT_INVOKER_POST_PARSE_ARGS)
            from knack.util import CtxTypeError

            def on_global_arguments(cli_ctx, **kwargs):
                arg_group = kwargs.get('arg_group')
                arg_group.add_argument(
                    '--controller-context',
                    dest='_context',
                    help='Controller context',
                    type=str.lower,
                )
                cli_ctx.global_context_arg.arg_group = arg_group

            def handle_context_parameter(cli_ctx, **kwargs):
                args = kwargs['args']
                # disable=protected-access
                context = args._context  # pylint:
                if context:
                    Configuration().context = context

            if cli_ctx is not None and not isinstance(cli_ctx, knack.CLI):
                raise CtxTypeError(cli_ctx)

            self.cli_ctx = cli_ctx
            self.arg_group = None
            self.cli_ctx.register_event(
                EVENT_PARSER_GLOBAL_CREATE,
                on_global_arguments
            )
            self.cli_ctx.register_event(
                EVENT_INVOKER_POST_PARSE_ARGS,
                handle_context_parameter
            )

        def remove_context_argument(self):
            if self.arg_group:
                for option in ['--context']:
                    for action in self.arg_group._actions:
                        #if vars(action)['option_strings']:
                        if vars(action)['option_strings'][0] == option:
                            self.arg_group._handle_conflict_resolve(None, [
                            (option, action)])
                            break
                self.arg_group = None

    class CLI(knack.CLI):
        """
        Extend CLI to override get_cli_version.
        """

        def __init__(self, **kwargs):
            self._supplied_file_object = None
            super(CLI, self).__init__(**kwargs)
            self.global_context_arg = GlobalContextArgument(cli_ctx=self)

        def get_cli_version(self):
            return '{0}\n{1}\n{2}\n{3}'.format(
                config.version,
                '\nBuild (' + config.build + ')\n' if config.build else '',
                'SQL Server 2019 ({0})'.format(config.server_version),
                '\nLegal docs and information: {0}'.format(config.LICENSE)
            )

        def exception_handler(self, ex):
            """
            The default exception handler
            """

            self.logging.add_error_handler()
            ex = self.output.error(ex, self.invocation.data['output'])

            # This allows us to print stderr to output-files, helpful in a
            # single process mode such as testing.
            if self._supplied_file_object:
                print(ex, file=self._supplied_file_object, end='')

            return 1

        def invoke(self, args, initial_invocation_data=None, out_file=None):
            """
            Invoke a command.

            :param args: The arguments that represent the command
            :type args: list, tuple
            :param initial_invocation_data: Prime the in memory collection of
                                            key-value data for this invocation.
            :type initial_invocation_data: dict
            :param out_file: The file to send output to. If not used, we use
                             out_file for knack.cli.CLI instance
            :type out_file: file-like object
            :return: The exit code of the invocation
            :rtype: int
            """
            from azdata.cli.core.output import OutputStream
            from azdata.cli.core.events import (
                EVENT_INVOKER_PRE_CMD_TBL_TRUNCATE,
                EVENT_CMD_OUTPUT
            )
            from knack.util import CommandResultItem
            from knack.events import (
                EVENT_CLI_PRE_EXECUTE,
                EVENT_CLI_POST_EXECUTE
            )

            self._supplied_file_object = out_file

            if not isinstance(args, (list, tuple)):
                raise TypeError('args should be a list or tuple.')
            exit_code = 0

            try:
                args = self.completion.get_completion_args() or args
                out_file = out_file or self.out_file
                output = '--output' in args or '-o' in args
                h = '--help' in args or '-h' in args

                # -- turn off console logging to stdout and do file only --
                # order of these two lines are important
                OutputProducer.structured_output = output and not h
                OutputStream().repurpose()

                # -- indicate global arguments are set for easy lookup --
                Configuration().verbose = '--verbose' in args
                Configuration().debug = '--debug' in args
                Configuration().interactive = False if output else \
                    Configuration().interactive

                # -- order matters --
                self.logging.configure(args)
                self.raise_event(EVENT_CMD_OUTPUT)
                self.raise_event(EVENT_CLI_PRE_EXECUTE)

                if CLI._should_show_version(args):
                    self.show_version()
                    self.result = CommandResultItem(None)
                else:
                    self.invocation = self.invocation_cls(
                        cli_ctx=self,
                        parser_cls=self.parser_cls,
                        commands_loader_cls=self.commands_loader_cls,
                        help_cls=self.help_cls,
                        initial_data=initial_invocation_data
                    )

                    self.raise_event(EVENT_INVOKER_PRE_CMD_TBL_TRUNCATE,
                                     args=args)

                    cmd_result = self.invocation.execute(args)
                    self.result = cmd_result
                    exit_code = self.result.exit_code
                    output_type = self.invocation.data['output']

                    if output or cmd_result and cmd_result.result is not None:
                        formatter = self.output.get_formatter(output_type)
                        self.output.out(
                            cmd_result,
                            formatter=formatter,
                            out_file=out_file
                        )

                self.raise_event(EVENT_CLI_POST_EXECUTE)
            except KeyboardInterrupt as ex:
                self.result = CommandResultItem(None, error=ex)
                exit_code = 1
            except Exception as ex:  # pylint: disable=broad-except
                exit_code = self.exception_handler(ex)
                self.result = CommandResultItem(None, error=ex)
            finally:
                pass

            self.result.exit_code = exit_code
            return exit_code

    return CLI(
        config_dir=config.CLI_CONFIG_DIR,
        cli_name=config.CLI_NAME,
        commands_loader_cls=CliCommandsLoader,
        help_cls=CliHelp,
        logging_cls=CliLogging,
        config_cls=CliConfig,
        output_cls=OutputProducer
    )
