# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------
from typing import Union
from pkgutil import get_loader
import os


class Paths(object):
    # The possible extensions that init files can have
    _init_extensions = [".py", ".pyc"]

    # The name of init files that live at module roots
    _init_file_name = "__init__"

    @staticmethod
    def core():
        """
        Gives the absolute path on the system of the root of the azdata.cli.core module.

        :returns: The absolute path on the system of azdata.cli.core module.
        """
        return Paths._resolve_module_path("azdata.cli.core")

    @staticmethod
    def core_file(relative_path: Union[list]):
        """
        Given a file path relative to the root of the core module, resolves the full path of
        that file on the system.

        :param relative_path: The path from the root of the core module provided as a list of strings.
        :returns: The full path on the system of the requested file.

        Examples:

        core_file(['templates', 'template.yaml']) on a linux system dev environment yields:
        "<project_root>/projects/cli/src/azdata-cli-core/azdata/cli/core/templates/template.yaml"

        core_file(['models', 'template.yaml'] on a windows system dev environment yields:
        "<project_root>\projects\cli\src\azdata-cli-core\azdata\cli\core\templates\template.yaml"
        """
        base = Paths.core()
        return os.path.abspath(os.path.join(base, *relative_path))

    @staticmethod
    def raw_module(module: str):
        """
        Resolves the root of any module that is loadable on the current system.

        **IMPORTANT**: Don't use this if you're looking for a plugin outside of the one you're writing, or the
        core module. Core has it's own dedicated method for the purpose of abstraction, and if you're searching for a
        plugin module then you're probably doing something wrong as that violates the total isolation we want for plugins.

        This method should be used rarely if at all.
        """
        return Paths._resolve_module_path(module)

    @staticmethod
    def _resolve_module_path(module: str):
        try:
            loader = get_loader(module)
            if loader is None:
                # TODO(jakedern): replace this with ModuleNotFoundError which is only available in python 3.6 and up
                raise ImportError("Module '{}' could not be found by package loader.".format(module))
        except ImportError:
            raise

        # Python files can have both .py and .pyc extensions, so we need to check all possible extensions
        # for the __init__ file and remove that part of the path if any of them match.
        path = loader.get_filename()
        for ext in Paths._init_extensions:
            file = Paths._init_file_name + ext
            if path.endswith(file):
                return path[:-len(file)]

        return path
